package main

import (
	"errors"
	"fmt"
	"strings"

	p "code.ivysaur.me/plexus"
)

type plexusEntry struct {
	name   string
	insert []string
}

type PlexusVersion struct {
	rules []plexusEntry
}

func bnfNameToPlexusName(in string) string {
	return "RULE_" + strings.ToUpper(in)
}

func expect(mn *p.MatchNode, t p.MatchType, nChildren int) error {
	if mn.Type != t {
		return fmt.Errorf("Expected type(%d), got type(%d)", t, mn.Type)
	}

	if len(mn.Children) != nChildren {
		return fmt.Errorf("Expected %d children for type(%d), got %d", nChildren, t, len(mn.Children))
	}

	return nil
}

func (this *PlexusVersion) rulesForSaeChild(name string, i int, sae_child *p.MatchNode) ([]string, error) {
	switch sae_child.Type {

	case RULE_IDENTIFIER:
		return []string{fmt.Sprintf(`p.Ref(%s)`, bnfNameToPlexusName(string(sae_child.Children[0].LeafContent)))}, nil

	case RULE_QUOTED_LITERAL:
		literalContent := sae_child.Children[0].LeafContent[1 : len(sae_child.Children[0].LeafContent)-1] // strip single-quotes
		return []string{fmt.Sprintf("p.Literal(`%s`)", string(literalContent))}, nil

	case RULE_ARGLIST_NESTED:
		innerName := fmt.Sprintf("%s__PAREN%d", name, i)
		err := this.addExpressionFrom(innerName, sae_child.Children[1].Children)
		if err != nil {
			return nil, err
		}

		return []string{fmt.Sprintf("p.Ref(%s)", innerName)}, nil

	case RULE_ARGLIST_OPTIONAL:
		innerName := fmt.Sprintf("%s__SQPAREN%d", name, i)
		elms, err := this.getExpressionFrom(innerName, sae_child.Children[1].Children)
		if err != nil {
			return nil, err
		}

		elms = append([]string{"p.Ref(RT_RULE_OPTIONAL_REMAINING)"}, elms...)
		this.rules = append(this.rules, plexusEntry{innerName, elms})

		return []string{fmt.Sprintf("p.Ref(%s)", innerName)}, nil

	default:
		return nil, fmt.Errorf("Unexpected type(%d) as child of single-arg-element", sae_child.Type)

	}
}

func (this *PlexusVersion) container(name string, i int, sae_child *p.MatchNode, rule string) ([]string, error) {
	contentName := fmt.Sprintf("%s__REP%d_CONTENT", name, i)
	innerContent, err := this.rulesForSaeChild(contentName, i, sae_child)
	if err != nil {
		return nil, err
	}
	this.rules = append(this.rules, plexusEntry{contentName, innerContent})

	// create repeatName for a wrapper element to control
	repeatName := fmt.Sprintf("%s__REP%d_CONTAINER", name, i)

	this.rules = append(this.rules, plexusEntry{repeatName, []string{rule, fmt.Sprintf("p.Ref(%s)", contentName)}})

	return []string{fmt.Sprintf(`p.Ref(%s)`, repeatName)}, nil
}

func (this *PlexusVersion) addExpressionFrom(name string, arglist []p.MatchNode) error {
	insert, err := this.getExpressionFrom(name, arglist)
	if err != nil {
		return err
	}

	this.rules = append(this.rules, plexusEntry{name, insert})
	return nil
}

func (this *PlexusVersion) getExpressionFrom(name string, arglist []p.MatchNode) ([]string, error) {
	insert := make([]string, 0)

	// Pass 1: If this arglist contains any RULE_OR, split into multiple children for p.RT_RULE_ONE_OF_THE_REMAINING
	or_base := 0
	for i := 0; i < len(arglist); i += 1 {
		err := expect(&arglist[i], RULE_SINGLE_ARG_ELEMENT, 1)
		if err != nil {
			return nil, err
		}

		sae_child := arglist[i].Children[0]

		if sae_child.Type == RULE_OR {
			err := this.addExpressionFrom(fmt.Sprintf("%s__OPT%d", name, or_base), arglist[or_base:i])
			if err != nil {
				return nil, err
			}
			insert = append(insert, fmt.Sprintf("p.Ref(%s__OPT%d)", name, or_base))
			or_base = i + 1
		}
	}
	if or_base > 0 {
		err := this.addExpressionFrom(fmt.Sprintf("%s__OPT%d", name, or_base), arglist[or_base:])
		if err != nil {
			return nil, err
		}

		insert = append([]string{"p.Ref(p.RT_RULE_ONE_OF_THE_REMAINING)"}, insert...)
		insert = append(insert, fmt.Sprintf("p.Ref(%s__OPT%d)", name, or_base))
		return insert, nil
	}

	// Pass 2: There are no RULE_OR in here - so they must all be RULE_SINGLE_ARG_ELEMENT_REPEATABLE.
	// Walk elements directly, handling LITERAL, REFERENCE, [], (), and ()* etc
	for i, sae := range arglist {

		sae_repeatable_container := sae.Children[0]
		sae_basic_container := sae_repeatable_container.Children[0] // safe, from the previous pass
		sae_child := sae_basic_container.Children[0]

		switch len(sae_repeatable_container.Children) {
		case 1:
			// normal
		case 2:
			// there are repeatflags on this sae_child
			flag_el := sae_repeatable_container.Children[1]
			if flag_el.LeafContent[0] == '+' {
				extra, err := this.container(name, i, &sae_child, "p.Ref(RT_RULE_REPEAT_NEXT_ONEPLUS)")
				if err != nil {
					return nil, err
				}
				insert = append(insert, extra...)

			} else if flag_el.LeafContent[0] == '*' {
				extra, err := this.container(name, i, &sae_child, "p.Ref(RT_RULE_REPEAT_NEXT_ZEROPLUS)")
				if err != nil {
					return nil, err
				}
				insert = append(insert, extra...)

			} else {
				return nil, errors.New("Unexpected repeatable type")
			}

		default:
			return nil, fmt.Errorf("Expected 1/2 children for sae-repeatable, got %d", len(sae_repeatable_container.Children))
		}

		// Inner sae_child

		str, err := this.rulesForSaeChild(name, i, &sae_child)
		if err != nil {
			return nil, err
		}

		insert = append(insert, str...)
	}

	// Done
	return insert, nil
}

func (this *PlexusVersion) addExpression(mn *p.MatchNode) error {
	err := expect(mn, RULE_EXPRESSION, 2)
	if err != nil {
		return err
	}

	name := bnfNameToPlexusName(string(mn.Children[0].Children[0].LeafContent))

	return this.addExpressionFrom(name, mn.Children[1].Children)
}

func Convert(mn *p.MatchNode) (*PlexusVersion, error) {
	if mn.Type != RULE_DOCUMENT {
		return nil, errors.New("Expected document root")
	}

	comment := ""
	ret := PlexusVersion{
		rules: make([]plexusEntry, 0),
	}

	for _, docEl := range mn.Children {
		err := expect(&docEl, RULE_DOCUMENT_ENTRY, 1)
		if err != nil {
			return nil, err
		}

		if docEl.Children[0].Type == RULE_COMMENT_LINE {
			comment = comment + string(docEl.Children[0].LeafContent)

		} else { // RULE_EXPRESSION
			err := ret.addExpression(&docEl.Children[0])
			if err != nil {
				return nil, err
			}
		}
	}

	return &ret, nil
}
