package plexus

import (
	"bytes"
	"errors"
	"fmt"
	"regexp"
)

type MatchType int

const (
	RT_LITERAL                    MatchType = -1
	RT_REGEXP                     MatchType = -2
	RT_RULE_ALL_IN_ORDER          MatchType = -3
	RT_RULE_OPTIONAL_REMAINING    MatchType = -4
	RT_RULE_ONE_OF_THE_REMAINING  MatchType = -5
	RT_RULE_REPEAT_NEXT_ONEPLUS   MatchType = -6
	RT_RULE_REPEAT_NEXT_ZEROPLUS  MatchType = -7
	RT_WHITESPACE_MODE            MatchType = -8
	RT_RULE_NONE_OF_THE_REMAINING MatchType = -9
)

func (mt MatchType) IsRule() bool {
	switch mt {
	case
		RT_RULE_ALL_IN_ORDER,
		RT_RULE_ONE_OF_THE_REMAINING,
		RT_RULE_OPTIONAL_REMAINING,
		RT_RULE_REPEAT_NEXT_ONEPLUS,
		RT_RULE_REPEAT_NEXT_ZEROPLUS,
		RT_RULE_NONE_OF_THE_REMAINING:
		return true
	}
	return false
}

type MatchNode struct {
	Type        MatchType   `json:"type"`
	Children    []MatchNode `json:"children,omitempty"`
	LeafContent []byte      `json:"content,omitempty"`
}

func (this *MatchNode) Equals(other *MatchNode) bool {
	if this.Type != other.Type {
		return false
	}
	if !bytes.Equal(this.LeafContent, other.LeafContent) {
		return false
	}
	if len(this.Children) != len(other.Children) {
		return false
	}
	for i := 0; i < len(this.Children); i += 1 {
		if !this.Children[i].Equals(&other.Children[i]) {
			return false
		}
	}
	return true
}

func Literal(content string) MatchNode {
	return MatchNode{Type: RT_LITERAL, LeafContent: []byte(content)}
}

func Regexp(content string) MatchNode {
	return MatchNode{Type: RT_REGEXP, LeafContent: []byte(content)}
}

func SkipWhitespace(cutset string) MatchNode {
	return MatchNode{Type: RT_WHITESPACE_MODE, LeafContent: []byte(cutset)}
}

func PreserveWhitespace() MatchNode {
	return MatchNode{Type: RT_WHITESPACE_MODE, LeafContent: []byte{}}
}

func Ref(t MatchType) MatchNode {
	return MatchNode{Type: t}
}

type SyntaxTree [][]MatchNode

type ParseAttempt struct {
	Syntax     SyntaxTree
	regexCache map[string]*regexp.Regexp
}

func NewParseAttempt() *ParseAttempt {
	return &ParseAttempt{
		Syntax:     make(SyntaxTree, 0),
		regexCache: make(map[string]*regexp.Regexp, 0),
	}
}

// Parse attempts to match a specific rule at the current position. If a match
// was found, returns the remaining content and a matched node, else nil+nil.
func (this *ParseAttempt) parseAt(parseRule MatchNode, whitespaceChars string, pos []byte) ([]byte, *MatchNode) {

	if len(whitespaceChars) > 0 {
		pos = bytes.TrimLeft(pos, whitespaceChars)
	}

	if len(pos) == 0 {
		return nil, nil
	}

	// Try to parse this token at this position
	switch parseRule.Type {

	case RT_LITERAL:
		if bytes.HasPrefix(pos, parseRule.LeafContent) {
			return pos[len(parseRule.LeafContent):], &MatchNode{Type: RT_LITERAL, LeafContent: parseRule.LeafContent}
		} else {
			return nil, nil
		}

	case RT_REGEXP:
		rx, ok := this.regexCache[string(parseRule.LeafContent)]
		if !ok {
			rx = regexp.MustCompilePOSIX("^" + string(parseRule.LeafContent))
			this.regexCache[string(parseRule.LeafContent)] = rx
		}

		loc := rx.FindIndex(pos)
		if loc == nil {
			return nil, nil // no match
		}
		if loc[0] != 0 {
			return nil, nil // found match not anchored at ^
		}
		return pos[loc[1]:], &MatchNode{Type: RT_LITERAL, LeafContent: pos[:loc[1]]}

	default:
		if int(parseRule.Type) >= len(this.Syntax) {
			// No built-in or external matcher for this type
			return nil, nil
		}
		knownRule := this.Syntax[parseRule.Type]

		// Composite matcher
		lastRule := RT_RULE_ALL_IN_ORDER
		elements := make([]MatchNode, 0, len(knownRule))
		repNextOnePlusHasMatched := false
	childRules:
		for _, el := range knownRule {
		repeatRule:
			for {

				if el.Type == RT_WHITESPACE_MODE {
					whitespaceChars = string(el.LeafContent)
					continue childRules

				} else if el.Type.IsRule() {
					lastRule = el.Type
					repNextOnePlusHasMatched = false

					if lastRule == RT_RULE_REPEAT_NEXT_ZEROPLUS {
						// Same as oneplus but with already accepting the end condition
						lastRule = RT_RULE_REPEAT_NEXT_ONEPLUS
						repNextOnePlusHasMatched = true
					}
					continue childRules
				}

				// Attempt recursive parse
				rpos, rmn := this.parseAt(el, whitespaceChars, pos)

				switch lastRule {
				case RT_RULE_ALL_IN_ORDER:
					// If it parsed, store it; otherwise, fail
					if rmn != nil {
						elements = append(elements, *rmn)
						pos = rpos
					} else {
						return nil, nil // couldn't parse mandatory object at this position
					}

				case RT_RULE_OPTIONAL_REMAINING:
					// Either all the remaining elements must occur, or none can.
					if rmn != nil {
						lastRule = RT_RULE_ALL_IN_ORDER // All remaining elements are required.
						elements = append(elements, *rmn)
						pos = rpos
					} else {
						break childRules // Success
					}

				case RT_RULE_ONE_OF_THE_REMAINING:
					if rmn != nil {
						lastRule = RT_RULE_ALL_IN_ORDER // don't hit later failure condition
						elements = append(elements, *rmn)
						pos = rpos
						break childRules
					}

				case RT_RULE_NONE_OF_THE_REMAINING:
					if rmn != nil {
						return nil, nil // parsed forbidden object at this position
					}

				case RT_RULE_REPEAT_NEXT_ONEPLUS:
					if rmn != nil {
						elements = append(elements, *rmn)
						pos = rpos
						repNextOnePlusHasMatched = true
						continue repeatRule

					} else {
						if repNextOnePlusHasMatched {
							// Failure okay, we got our one+
							break childRules
						} else {
							return nil, nil // couldn't parse one+ of these
						}
					}

				default:
					panic("Unexpected last rule")
				}

				break repeatRule // normally don't repeat
			}
		}

		if lastRule == RT_RULE_ONE_OF_THE_REMAINING {
			// didn't match remaining rule
			return nil, nil
		} else {
			// made it through all match cases
			return pos, &MatchNode{Type: parseRule.Type, Children: elements}
		}
	}

}

type ParseFailure struct {
	CharPos, LineNo, ColumnNo int
}

func (pf ParseFailure) Error() string {
	return fmt.Sprintf("Parse failure on line %d", pf.LineNo)
}

func lookupIndex(index int, content []byte) (linect, offset int) {
	linect = 1
	last_line := 0
	for i := 0; i < index; i += 1 {
		if content[i] == '\n' {
			linect += 1
			last_line = i
		}
	}
	offset = index - last_line
	return
}

func (this *ParseAttempt) Parse(parseAsType MatchType, content []byte) (*MatchNode, error) {
	defaultWhitespaceChars := " \t\r\n"

	remaining, mn := this.parseAt(Ref(parseAsType), defaultWhitespaceChars, content)
	if mn == nil {
		return nil, errors.New("Couldn't parse input")
	}

	// Skip over any trailing whitespace
	extra := bytes.TrimLeft(remaining, defaultWhitespaceChars)
	if len(extra) > 0 {
		failPos := len(content) - len(remaining)
		linect, offset := lookupIndex(failPos, content)
		return nil, ParseFailure{CharPos: failPos, LineNo: linect, ColumnNo: offset}
	}

	return mn, nil
}
